let bridge, ip, user, hueLightId
let watchRay = [1, 0, 0]
let lastTap

let calibratingLightId = null
let lightDirections = {}
let lightStates = {}
let lightBrightnesses = {}
let computerDirection = null
let calibratingComputer = false
let computerHovering = false

let cursorX, cursorY
let cumulativeHandRotation = 0
let handRotationOrigin
let rotatingLight = false
let rotatingStarted = false

const getURLParameters = () => {
    const query = window.location.search.substring(1)
    const params = new URLSearchParams(query)
    return Object.fromEntries(params.entries())
}

const setURLParameters = (parameters, reload) => {
    const params = new URLSearchParams(parameters)

    if (reload) {
        // Causes page refresh
        window.location.search = params.toString()
    } else {
        // Doesn't cause page refresh
        window.history.replaceState({}, '', `${location.pathname}?${params.toString()}`)
    }
}

const createConnectButton = (ip_) => {
    const connectButton = document.createElement('button')
    connectButton.innerText = `Connect to ${ip_}`
    connectButton.addEventListener('click', () => {
        // Create new user
        newUser(ip_).then(result => {
            if (result[0].error && result[0].error.type === 101) {
                alert('Press Link button on the Hue Bridge, then press the Connect button.')
            } else if (result[0].success) {
                const user_ = result[0].success.username
                connectButton.classList.add('hidden')
                setURLParameters({
                    'ip': ip_,
                    'user': user_
                }, true)
            } else {
                console.log('Unknown response from Hue bridge:', result)
            }
        })
    })
    return connectButton
}

const updateLightsTable = (readLightStates) => {
    calibratingLightId = null
    calibratingComputer = false

    bridge.getLights().then(lights => {
        const lightListElement = document.querySelector('#lights-list')
        lightListElement.innerHTML = ''

        Object.entries(lights).forEach(([lightId, light]) => {
            if (light.state.reachable) {
                if (readLightStates !== false) {
                    lightStates[lightId] = light.state.on && light.state.bri > 50
                    if (light.state.bri > 50) {
                        lightBrightnesses[lightId] = light.state.bri
                    } else {
                        lightBrightnesses[lightId] = 254
                    }
                }
 
                const row = document.createElement('tr')

                let directionText = ''
                const dir = lightDirections[lightId]
                if (dir) {
                    const x = dir[0].toFixed(3)
                    const y = dir[1].toFixed(3)
                    const z = dir[2].toFixed(3)
                    directionText = `${x}, ${y}, ${z}`
                }

                row.innerHTML = `
                <td>${light.name}</td>
                <td class="status">${light.state.on ? 'ON' : 'OFF'}</td>
                <td><button class="on">On</button></td>
                <td><button class="off">Off</button></td>
                <td><button class="calibrate">Calibrate</button></td>
                <td>${directionText}</td>
                <td><span id="lamp-${lightId}-hover" class="hidden">hover</span></td>
                `

                row.querySelector('button.on').addEventListener('click', () => {
                    setLightOnOffAndHover(lightId, true, false).then(() => {
                        updateLightsTable()
                    })
                })

                row.querySelector('button.off').addEventListener('click', () => {
                    setLightOnOffAndHover(lightId, false, false).then(() => {
                        updateLightsTable()
                    })
                })

                const calibrateButton = row.querySelector('button.calibrate')
                calibrateButton.addEventListener('click', () => {
                    calibratingLightId = lightId
                    setLightOnOffAndHover(lightId, false, true)
                    calibrateButton.parentNode.innerHTML = 'Tap to set direction'
                })

                lightListElement.appendChild(row)
            }
        })

        const row = document.createElement('tr')

        let directionText = ''
        const dir = computerDirection
        if (dir) {
            const x = dir[0].toFixed(3)
            const y = dir[1].toFixed(3)
            const z = dir[2].toFixed(3)
            directionText = `${x}, ${y}, ${z}`
        }

        row.innerHTML = `
        <td>Computer</td>
        <td></td>
        <td></td>
        <td></td>
        <td><button class="calibrate">Calibrate</button></td>
        <td>${directionText}</td>
        <td><span id="computer-hover" class="hidden">hover</span></td>
        `

        const calibrateButton = row.querySelector('button.calibrate')
        calibrateButton.addEventListener('click', () => {
            calibratingComputer = true
            computerHoverEffect(true)
            calibrateButton.parentNode.innerHTML = 'Tap to set direction'
        })

        lightListElement.appendChild(row)
    })
}

const computerHoverEffect = (value) => {
    const badgeElement = document.querySelector(`#computer-hover`)
    if (value) {
        badgeElement.classList.remove('hidden')
    } else {
        badgeElement.classList.add('hidden')
    }
}


let lightsHoveredOn = {}

const setLightHoverState = (lightId, visible) => {
    const badgeElement = document.querySelector(`#lamp-${lightId}-hover`)
    if (visible) {
        if (!lightsHoveredOn[lightId]) {
            setLightOnOffAndHover(lightId, lightStates[lightId], true, 0)
        }
        lightsHoveredOn[lightId] = true
        badgeElement.classList.remove('hidden')
    } else {
        if (lightsHoveredOn[lightId]) {
            setLightOnOffAndHover(lightId, lightStates[lightId], false, 0)
        }
        lightsHoveredOn[lightId] = false
        badgeElement.classList.add('hidden')
    }
}

const setLightOnOffAndHover = (lightId, onoff, hover, transitiontime) => {
    if (transitiontime === undefined) {
        transitiontime = 6
    }
    if (onoff) {
        if (hover) {
            return bridge.setLightState(lightId, {
                'on': true,
                'bri': lightBrightnesses[lightId],
                'hue': 8414,//Math.floor(65536./360*230),
                'sat': 140,
                'transitiontime': transitiontime
            })
        } else {
            return bridge.setLightState(lightId, {
                'on': true,
                'bri': Math.floor(240/254 * lightBrightnesses[lightId]),
                'hue': 8414,
                'sat': 140,
                'transitiontime': transitiontime
            })
        }
    } else {
        if (hover) {
            return bridge.setLightState(lightId, {
                'on': true,
                'bri': 15,
                'hue': Math.floor(65536./360*10),
                'sat': 254,
                'transitiontime': transitiontime
            })
        } else {
            return bridge.setLightState(lightId, {
                'on': false,
                'transitiontime': transitiontime
            })
        }
    }
}

const setlightBrightness = (lightId, brightness) => {
    brightness = Math.max(0, Math.min(254, brightness))
    brightness = Math.floor(brightness)

    lightBrightnesses[lightId] = brightness

    return bridge.setLightState(lightId, {
        'on': true,
        'bri': brightness,
        'hue': 8414,
        'sat': 140,
        'transitiontime': 1
    })
}

const setVirtualCursorPosition = (cursorX, cursorY) => {
    const cursor = document.querySelector('#cursor')
    cursor.style.left = `${cursorX}px`
    cursor.style.top = `${cursorY}px`
}

const showVirtualCursor = () => {
    const cursor = document.querySelector('#cursor')
    cursor.classList.remove('hidden')
    document.body.style.cursor = 'none'
}

const showRealCursor = () => {
    const cursor = document.querySelector('#cursor')
    cursor.classList.add('hidden')
    document.body.style.cursor = 'auto'
}


window.addEventListener('load', () => {

    const bridgeListElement = document.querySelector('#hue-bridge-list')
    const discoverButton = document.querySelector('#discover-hue-bridge-button')

    const parameters = getURLParameters()

    ip = parameters.ip
    user = parameters.user
    console.log('ip', ip)
    console.log('user', user)

    if (ip === undefined) {
        discoverButton.classList.remove('hidden')
        discoverButton.addEventListener('click', () => {
            discoverBridges().then(bridgeList => {
                bridgeListElement.innerHTML = ''
                bridgeListElement.classList.remove('hidden')
                discoverButton.classList.add('hidden')

                bridgeList.forEach(bridge => {
                    const ip_ = bridge.internalipaddress

                    setURLParameters({
                        'ip': ip_
                    }, false)

                    const connectButton = createConnectButton(ip_)
                    bridgeListElement.appendChild(connectButton)
                })
            })
        })

    } else if (user === undefined) {
        bridgeListElement.innerHTML = ''
        bridgeListElement.classList.remove('hidden')

        const connectButton = createConnectButton(ip)
        bridgeListElement.appendChild(connectButton)

    } else {
        bridge = new HueBridge(ip, user)

        updateLightsTable()
    }



    const canvas = document.querySelector('#watch-canvas')
    const ctx = canvas.getContext('2d')
    const devicePixelRatio = window.devicePixelRatio || 1
    canvas.width = canvas.clientWidth * devicePixelRatio
    canvas.height = canvas.clientHeight * devicePixelRatio
    ctx.scale(devicePixelRatio, devicePixelRatio)
    const drawWidth = canvas.width / devicePixelRatio
    const drawHeight = canvas.height / devicePixelRatio
    canvas.classList.add('hidden')

    const drawGraphics = () => {
        const getCSSVar = (name) => window.getComputedStyle(document.body).getPropertyValue(name)

        // watch edge
        ctx.strokeStyle = getCSSVar('--fg-color')
        ctx.lineWidth = 30
        ctx.beginPath()
        ctx.arc(0, 0, 950, 0, 2 * Math.PI)
        ctx.stroke()

        // watch direction
        const x = -watchRay[1] * 880
        const y = -watchRay[2] * 880

        // moving line
        ctx.beginPath()
        ctx.moveTo(0, 0)
        ctx.lineTo(x, y)
        ctx.stroke()

        // small moving dot
        ctx.fillStyle = getCSSVar('--bg-color')
        ctx.beginPath()
        ctx.arc(x, y, 50, 0, 2 * Math.PI)
        ctx.fill()
        ctx.stroke()

        // tap circle
        if (lastTap !== undefined) {
            const tapAge = new Date() - lastTap
            transparency = Math.min(1, Math.max(0, tapAge / 300))
            ctx.globalAlpha = 1 - transparency
    
            ctx.fillStyle = getCSSVar('--fg-color')
            ctx.beginPath()
            ctx.arc(0, 0, 800, 0, 2 * Math.PI)
            ctx.fill()

            ctx.globalAlpha = 1
        }
    }

    const step = () => {
        ctx.clearRect(0, 0, canvas.width, canvas.height)
        ctx.fillStyle = 'hsla(0, 0%, 0%, 0%)'
        ctx.fillRect(0, 0, canvas.width, canvas.height)
        ctx.fill()

        const old = ctx.getTransform()
        ctx.translate(drawWidth / 2, drawHeight / 2) // (0, 0) = center
        ctx.scale(drawHeight / 2000, drawHeight / 2000) // height = -1000..1000

        drawGraphics()
    
        ctx.setTransform(old)
        ctx.restore()
        window.requestAnimationFrame(step)
    }
    window.requestAnimationFrame(step)

    let x = 0
    let y = 0
    let gravityDirection = {
        x: 0,
        y: 0,
        z: 1
    }
    let timeSincePreviousLightnessUpdate = new Date()

    const button = document.querySelector('#establish-connection')

    button.addEventListener('click', () => {

        if (navigator.bluetooth) {

            getWatch().then(watch => {

                watch.addEventListener('tap', () => {
                    console.log('tap')
                    lastTap = new Date()

                    if (calibratingLightId !== null) {
                        lightDirections[calibratingLightId] = watchRay
                        calibratingLightId = null
                        setLightOnOffAndHover(calibratingLightId, lightStates[calibratingLightId], false)
                        updateLightsTable(false)
                    
                    } else if (calibratingComputer) {
                        computerDirection = watchRay
                        calibratingComputer = false
                        computerHoverEffect(false)
                        updateLightsTable(false)

                    } else if (computerHovering) {
                        const element = document.elementFromPoint(cursorX, cursorY)
                        if (element) {
                            element.click()
                        }
                    } else {
                        Object.entries(lightsHoveredOn).forEach(([lightId, hovering]) => {
                            if (hovering) {
                                lightStates[lightId] = !lightStates[lightId]
                                setLightOnOffAndHover(lightId, lightStates[lightId], true)

                                if (!rotatingLight) {
                                    // Inverse of Math.pow(1 + dz/40, 2) * 254
                                    const brightnessOrigin = (Math.pow(lightBrightnesses[lightId] / 254, 0.5) - 1)*40
                                    handRotationOrigin = cumulativeHandRotation - brightnessOrigin
                                }
                                rotatingLight = true
                            }
                        })
                    }
                })

                watch.addEventListener('orientationchanged', event => {
                    const Q = event.detail
                    const q0 = Q[0]
                    const q1 = Q[1]
                    const q2 = Q[2]
                    const q3 = Q[3]

                    // First row of the rotation matrix
                    rxx = 2 * (q0 * q0 + q1 * q1) - 1
                    rxy = 2 * (q1 * q2 - q0 * q3)
                    rxz = 2 * (q1 * q3 + q0 * q2)
                     
                    // Second row of the rotation matrix
                    ryx = 2 * (q1 * q2 + q0 * q3)
                    ryy = 2 * (q0 * q0 + q2 * q2) - 1
                    ryz = 2 * (q2 * q3 - q0 * q1)
                     
                    // Third row of the rotation matrix
                    rzx = 2 * (q1 * q3 - q0 * q2)
                    rzy = 2 * (q2 * q3 + q0 * q1)
                    rzz = 2 * (q0 * q0 + q3 * q3) - 1

                    watchRay = [rzx, rzy, rzz]
                    const dot = (a, b) => a[0]*b[0] + a[1]*b[1] + a[2]*b[2]

                    const doHoverEnterHaptics = () => {
                        watch.triggerHaptics(1, 10)
                    }

                    const doHoverExitHaptics = () => {
                        watch.triggerHaptics(0.5, 10)
                    }

                    Object.entries(lightDirections).forEach(([lightId, direction]) => {
                        if (calibratingLightId === lightId) {
                            return
                        }
                        let threshold = 0.92
                        let wasHovering = false
                        if (lightsHoveredOn[lightId]) {
                            threshold = 0.9
                            wasHovering = true
                            if (rotatingStarted) {
                                threshold = 0.85
                            }
                        }
                        if (dot(watchRay, direction) > threshold) {
                            // allow only one light to be hovered on at once
                            if (Object.values(lightsHoveredOn).indexOf(true) == -1) {
                                setLightHoverState(lightId, true)
                                if (!wasHovering) {
                                    doHoverEnterHaptics()
                                }    
                            }
                        } else {
                            setLightHoverState(lightId, false)
                            if (wasHovering) {
                                doHoverExitHaptics()
                                rotatingLight = false
                                rotatingStarted = false
                            }
                        }
                    })

                    if (computerDirection) {
                        let threshold = 0.92
                        let wasHovering = false
                        if (computerHovering) {
                            threshold = 0.9
                            wasHovering = true
                        }
                        if (dot(watchRay, computerDirection) > threshold) {
                            computerHovering = true
                            computerHoverEffect(true)
                            if (!wasHovering) {
                                doHoverEnterHaptics()
                            }
                        } else {
                            computerHovering = false
                            computerHoverEffect(false)
                            if (wasHovering) {
                                doHoverExitHaptics()
                            }
                        }
                    }
                })

                watch.addEventListener('angularvelocitychanged', event => {
                    if (calibratingLightId !== null || calibratingComputer) {
                        return
                    }

                    const angularVelocity = event.detail
                    const vx = -angularVelocity[2] // hand right = +
                    const vy = -angularVelocity[1] // hand down = +
                    const vz = -angularVelocity[0] // hand clock-wise = +

                    // Rotate to change lightness
                    cumulativeHandRotation += vz

                    // 30~40 rotation units = 100% lightness change
                    Object.entries(lightsHoveredOn).forEach(([lightId, hovering]) => {
                        if (hovering) {
                            const dz = cumulativeHandRotation - handRotationOrigin
                            if (!dz) return
                            if (!rotatingLight) return
                            if (!lightStates[lightId]) return

                            if (Math.abs(dz) > 5 && !rotatingStarted) {
                                rotatingStarted = true
                                console.log('increasing cone')
                                // if (dz < -5) {
                                //     dz += 5
                                //     handRotationOrigin -= 5
                                // } else if (dz > 5) {
                                //     dz -= 5
                                //     handRotationOrigin += 5
                                // }
                            }

                            // there is some kind of bug related to precise starting in the code above ^
                            // so instead we use rotatingStarted only for increasing the
                            // raycasting cone after a bit of rotation (5 units), not
                            // actually disabling brightness adjustment before the hand
                            // has rotated a certain amount.
                            //
                            // the problem is repeated on/offs change the brightness even
                            // if the hand doesn't rotate.
                            //
                            // thus, this line is commented.
                            // if (!rotatingStarted) return

                            const now = new Date()
                            if (now - timeSincePreviousLightnessUpdate > 100) {
                                timeSincePreviousLightnessUpdate = now
                                let dzNorm = 1 + dz/40
                                dzNorm = Math.max(0, Math.min(1, dzNorm))
                                const brightness = Math.pow(dzNorm, 2) * 254
                                // console.log(brightness)
                                setlightBrightness(lightId, brightness)
                            }
                        }
                    })

                    // Ray cursor
                    if (computerHovering) {
                        const vr = Math.sqrt(vx*vx + vy*vy)

                        const acceleration = 0.10
                        const speed = 70
                        const dx = vx * speed * Math.pow(vr, acceleration)
                        const dy = vy * speed * Math.pow(vr, acceleration)
    
                        cursorX += dx * gravityDirection.z + dy * gravityDirection.y
                        cursorY += dy * gravityDirection.z - dx * gravityDirection.y
    
                        cursorX = Math.min(window.innerWidth, Math.max(0, cursorX))
                        cursorY = Math.min(window.innerHeight, Math.max(0, cursorY))
    
                        if (vr > 0.05) {
                            showVirtualCursor()
                        }
                        setVirtualCursorPosition(cursorX, cursorY)    
                    }
                })

                watch.addEventListener('gravityvectorchanged', event => {
                    const gravityVector = event.detail
                    const x = gravityVector[0]
                    const y = gravityVector[1]
                    const z = gravityVector[2]
                    const r = Math.sqrt(x*x + y*y + z*z)
                    gravityDirection = {
                        x: x/r,
                        y: y/r,
                        z: z/r
                    }

                    // +z = oikeinpäin. pitää jakaa kympillä
                    // -y = y->-x & x->y
                })

                watch.addEventListener('disconnected', () => {
                    button.classList.remove('hidden')
                    canvas.classList.add('hidden')
                })

                button.classList.add('hidden')
                canvas.classList.remove('hidden')
            })

        } else {
            if (navigator.userAgent.indexOf("Chrome") != -1) {
                // Browser probably supports Web Bluetooth, but it is not enabled.
                alert('Web Bluetooth is not enabled. Please enable the'
                    +' "Experimental Web Platform features" flag on your browser.')
            } else {
                alert('Web Bluetooth is likely not supported on your browser.'
                    + ' Please try a Chrome-based browser.')
            }
        }
    })

    window.addEventListener('mousemove', (event) => {
        showRealCursor()

        cursorX = event.clientX
        cursorY = event.clientY
        setVirtualCursorPosition(cursorX, cursorY)
    })
    showRealCursor()
})